'''SSS Server Classes'''
__revision__ = '$Revision: 1.193 $'

from __future__ import nested_scopes

from atexit import register
from cPickle import loads, dumps, load, dump
from elementtree.ElementTree import Element, SubElement, XML, tostring
from fcntl import fcntl, F_SETFL
from os import getpid, kill, wait, write, read, O_NONBLOCK, pipe
from select import select, error as selecterror
from signal import signal, SIGINT, SIGTERM, SIGCHLD
from socket import gethostname, SOL_SOCKET, SO_REUSEADDR, AF_INET, SOCK_STREAM
import socket
from sys import exc_info
import sys
from syslog import openlog, syslog, closelog, LOG_INFO, LOG_ERR, LOG_LOCAL0
from time import localtime, strftime, sleep
from traceback import extract_tb
from types import ClassType

from sss.ssslib import comm_lib, ConnectError, CommunicationError, ProtocolError, SocketError, LookupError

class FakeValidationException:
    '''Used in cases where validation is not enabled'''
    pass

try:
    from validate import validate, ValidationException
except ImportError:
    ValidationException = FakeValidationException

class StateLoadError(Exception):
    '''Error in case of state load problems'''
    pass

class StateSaveError(Exception):
    '''Error in case of state save problems'''
    pass

class MsgError(Exception):
    '''Error for pipe framing cases'''
    pass

class BufferedPipe(object):
    '''Nonblocking pipe with attached buffer'''
    __separator__ = '|'
    def __init__(self):
        (self.r, self.w) = pipe()
        fcntl(self.w, F_SETFL, O_NONBLOCK)
        self.buffer = ''
        self.failed = False

    def WriteMessage(self, msg):
        '''frame and place into the buffer'''
        emsg = dumps(msg)
        self.buffer += "%s%16d%s" % (self.__separator__, len(emsg), emsg)

    def FlushMessages(self):
        '''write messages to pipe'''
        if not self.buffer:
            return
        for run in range(50):
            try:
                num = write(self.w, self.buffer)
                if self.failed:
                    syslog(LOG_ERR, "BP: Write OK")
                    self.failed = False
                self.buffer = self.buffer[num:]
            except OSError:
                if not self.failed:
                    syslog(LOG_ERR, "BP Write Failed")
                    self.failed = True
                    
    def ReadMessages(self):
        '''read messages from pipe'''
        self.buffer += read(self.r, 2048)

    def SkipBadMsg(self):
        if self.buffer[0] == self.__separator__:
            self.buffer = self.buffer[1:]
        syslog(LOG_ERR, "BP: Truncating buffer")
        hloc = self.buffer.find(self.__separator__)
        if hloc == -1:
            syslog(LOG_ERR, "BP: no new header found. flushing buffer contents")
            self.buffer = ''
        else:
            syslog(LOG_ERR, "BP: Skipping partial message %s bytes" % hloc)
            self.buffer = self.buffer[hloc:]

    def ReadMessage(self):
        '''de-frame messages from buffer'''
        # header = self.buffer[0]
        if self.buffer[0] != self.__separator__:
            self.SkipBadMsg()
        if len(self.buffer) < 17:
            syslog(LOG_ERR, "BP: Header incomplete. want:17 have %s" % (len(self.buffer)))
            syslog(LOG_ERR, self.buffer)
            self.SkipBadMsg()
            raise MsgError, "incomplete"
        try:
            msglen = int(self.buffer[1:17])
        except:
            self.SkipBadMsg()
        if len(self.buffer) < 17+msglen:
            syslog(LOG_ERR, "BP: Msg incomplete. want:%s have:%s" % (17+msglen, len(self.buffer)))
            raise MsgError, "incomplete"
        try:
            msg = loads(self.buffer[17:msglen+17])
        except:
            syslog(LOG_ERR, "BP: pickle")
            self.SkipBadMsg()
            raise MsgError, "pickle"
        self.buffer = self.buffer[msglen+17:]
        return msg

class SubscriptionException(Exception):
    '''Error for event recievers'''
    pass

class EventManagerError(Exception):
    '''Raised in cases of event propagation errrors'''
    pass

def AddEvent(component, msg, data):
    '''Cheesy function to send an event'''
    comm = comm_lib()
    handle = comm.ClientInit('event-manager')
    emsg = Element("add-event")
    SubElement(emsg, "event", component=component, msg=msg, data=data,
               time=strftime("%c", localtime()))
    comm.SendMessage(handle, tostring(emsg))
    response = comm.RecvMessage(handle)
    comm.ClientClose(handle)
    rxml = XML(response)
    if len(rxml.getchildren()) != 1:
        raise EventManagerError

def VPrint(results):
    '''Print column output'''
    lendata = [[len(res) for res in result] for result in results]
    fsizes = tuple([max([row[index] for row in lendata]) for index in range(len(lendata[0]))])
    fstring = "%%%ss   %%%ss   %%%ss" % fsizes
    print fstring % results[0]
    print 80 * "="
    for data in results:
        print fstring % data

class Server:
    '''This is a basic SSS Server class'''
    __component__ = 'BaseClass'
    __dispatch__ = {}
    __implementation__ = 'Server'
    __statefields__ = []
    __timeout__ = 10
    __validate__ = 1
    cfname = '/dev/null'
    
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.debug = kwargs.get('debug', 0)
        self.comm = comm_lib(debug=self.debug)
        self.pipes = []
        self.port = kwargs.get('port', 0)
        if (("validate" in globals().keys()) and self.__validate__):
            self.validate = 1
        else:
            self.validate = 0
        prefs = self.comm.GetConfig()
        if not prefs['prefs']:
            self.protocol = "challenge"
        else:
            self.protocol = kwargs.get('protocol', prefs['prefs'][0])
        self.__statefile__ = "/var/spool/sss/%s" % (self.__implementation__)
        self.schema_version = kwargs.get('schema_version', '1.0')
        self.tier = kwargs.get('tier', '1')
        self.__setup__()
        self.handles = {}
        self.peers = {}
        self.shut = False
        sys.excepthook = self.syslog_tb
        self.readsockets = []
        self.writesockets = []
        self.socketdispatch = {}
        registered = False
        for data in ['__implementation__', '__component__', '__dispatch__']:
            if not hasattr(self, data):
                raise TypeError, "%s not set" % (data)
        openlog(self.__implementation__, 0, LOG_LOCAL0)
        try:
            d = self.comm.StaticLookup(self.__component__)
            self.comm.SetServerProtocol(d['protocol'])
            self.port = int(d['extra'])
            syslog(LOG_INFO, "Got static port %s for service %s"%(self.port, self.__component__))
            self.static = True
        except LookupError:
            self.static = False
        # Load saved state
        if self.__statefields__:
            try:
                self.LoadState()
            except StateLoadError:
                syslog(LOG_ERR, "Error loading saved state")
        # setup child procs
        self.children = self.__setupChildren__()
        # setup signal handlers
        signal(SIGINT, self.SigHand)
        signal(SIGTERM, self.SigHand)
        signal(SIGCHLD, self.SigChildHand)
        self.server = socket.socket(AF_INET, SOCK_STREAM)
        self.server.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        try:
            self.server.bind(('', self.port))
        except socket.error:
            syslog(LOG_ERR, "Failed to bind to port %s" % self.port)
            self.shut = 1
            return False
        self.__postbind__()
        self.server.listen(32)
        self.port = self.server.getsockname()[1]
        syslog(LOG_INFO, "Bound to port %s" % (self.port))
        self.readsockets.append(self.server)
        self.socketdispatch[self.server] = self.AcceptNewConnection
        if self.__component__ != "service-directory":
            try:
                if not self.static:
                    if not self.shut:
                        for attempt in range(5):
                            try:
                                self.comm.SDRegister(self.__component__, gethostname(), self.port,
                                                     self.protocol, self.schema_version, self.tier)
                                registered = True
                                break
                            except ConnectError:
                                syslog(LOG_ERR, "Failed to connect to the service directory. attempt: %s" %
                                       attempt)
                                sleep(0.5)
                                continue
            except:
                self.log_failure("sdregister")
                self.shut = 1
            if not registered:
                self.shut = 1
                syslog(LOG_ERR, "Failed to register with the service directory; shutting down")
        if not hasattr(self, 'schemabase'):
            self.schemabase = self.__component__
        if registered:
            # use atexit.register for cleanup
            register(self.Shutdown)

    def __setup__(self):
        '''__setup__ initializes subclass location member data'''
        pass

    def __setupChildren__(self):
        '''__setupChildren__ forks children, returning their pids'''
        return []

    def __shutdown__(self):
        '''Local shutdown hook'''
        pass

    def __progress__(self):
        '''Steps to run between selects'''
        return 0

    def __postbind__(self):
        '''Steps to execute on startup after network init'''
        pass

    def syslog_tb(self, arg1, arg2, arg3):
        '''Log traceback'''
        for line in extract_tb(arg3):
            syslog(LOG_ERR, '  File "%s", line %i, in %s' % (line[0], line[1], line[2]))
            print '  File "%s", line %i, in %s' % (line[0], line[1], line[2])
            syslog(LOG_ERR, '    %s' % line[3])
            print '    %s' % line[3]
        syslog(LOG_ERR, "%s: %s" % (arg1, arg2))
        print "%s: %s" % (arg1, arg2)

    def log_failure(self, failure):
        '''Log Failures in unexpected cases'''
        (trace, val, trb) = exc_info()
        syslog(LOG_ERR, "Unexpected failure in %s"%(failure))
        for line in extract_tb(trb):
            syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % line)
        syslog(LOG_ERR, "%s: %s\n"%(trace, val))
        del trace, val, trb

    def AcceptNewConnection(self, sock):
        '''Initiate new tcp session'''
        (newsock, sockinfo) = sock.accept()
        try:
            self.handles[newsock] = self.comm.ServerInit(newsock)
            self.readsockets.append(newsock)
            self.peers[newsock] = sockinfo
        except (CommunicationError, ProtocolError, SocketError), msg:
            syslog(LOG_ERR, "Failed to init %s: %s" % (sockinfo[0], msg))
        return 0

    def ShutdownPeer(self, sock):
        '''Clean up after shutdown tcp session'''
        handle = self.handles[sock]
        self.readsockets.remove(sock)
        del self.handles[sock]
        del self.peers[sock]
        self.comm.ServerClose(handle)

    def Serve(self):
        '''Handle single select loop pass'''
        messages = 0
        try:
            (inl, outl, other) = select(self.readsockets, self.writesockets, [], self.__timeout__)
        except selecterror, serr:
            if serr[0] != 4:
                # ! EINTR
                # We got some other freaky error
                self.log_failure("select")
                syslog(LOG_ERR, str(self.readsockets))
                syslog(LOG_ERR, str(self.socketdispatch))
            return 0
        for sock in inl + outl:
            if self.socketdispatch.has_key(sock):
                try:
                    messages += self.socketdispatch[sock](sock)
                except:
                    self.log_failure("socket dispatch")
                    syslog(LOG_ERR, str(self.readsockets))
                    syslog(LOG_ERR, str(self.socketdispatch))
            else:
                messages += self.HandleSSSMessage(sock)
        return messages
                
    def ServeN(self, num):
        '''Process N requests'''
        done = 0
        while done < num:
            done += self.Serve()
            
    def ServeForever(self):
        '''Serve until shutdown flag is set'''
        unsaved = 0
        num = 0
        while not self.shut:
            try:
                num = self.Serve()
            except selecterror:
                num = 0
            try:
                num += self.__progress__()
            except:
                self.log_failure("progress")

            unsaved += num
                
            for p in self.pipes:
                p.FlushMessages()
            if self.__statefields__:
                if unsaved != 0:
                    try:
                        self.SaveState()
                    except StateSaveError:
                        syslog(LOG_ERR, "Error saving state; disabling data persistence")
                        self.__statefields__ = []
                    unsaved = 0

    def SigHand(self, signum, frame):
        '''Shutdown on unexpected signals'''
        self.shut = True

    def SigChildHand(self, signum, frame):
        '''Handle Child death'''
        try:
            (pid, status)=wait()
            syslog(LOG_INFO, "pid %s exited" % (pid))
            if pid in self.children:
                self.children.remove(pid)
        except OSError:
            pass
            
    def Shutdown(self):
        '''Shutdown and cleanup server'''
        self.__shutdown__()
        # Kill children
        [kill(pid, SIGINT) for pid in self.children]
        self.server.close()
        syslog(LOG_INFO, "Shutting down")
        if self.__statefields__:
            try:
                self.SaveState()
            except:
                syslog(LOG_ERR, "Error saving state")
        if self.__component__ != "service-directory":
            try:
                if not self.static:
                    self.comm.SDRemove(self.__component__, gethostname(), self.port, self.protocol)
            except:
                self.log_failure("sdremove")
        closelog()

    def HandleSSSMessage(self, sock):
        '''Handle sss message recv/send on sock'''
        messages = 0
        (peer, peerport) = self.peers[sock]
        try:
            msg = self.comm.RecvMessage(self.handles[sock])
            messages += 1
        except (CommunicationError, SocketError, ProtocolError), msg:
            syslog(LOG_ERR, "%s from %s" % (msg, peer))
            syslog(LOG_ERR, "closing socket")
            self.ShutdownPeer(sock)
            return 0

        if msg == '':
            self.ShutdownPeer(sock)
            return 0

        try:
            cmd = XML(msg)
        except:
            syslog(LOG_ERR, "parse error from %s: %s" % (peer, msg))
            self.comm.SendMessage(self.handles[sock], "<error type='Parse' msg='%s'/>" % msg)
            self.ShutdownPeer(sock)
            return 0

        if self.__dispatch__.has_key(cmd.tag):
            aname = self.__dispatch__[cmd.tag]
            oref = self
            while aname.count('.') > 0:
                (key, val) = aname.split('.', 1)
                oref = getattr(oref, key)
                aname = val
            func = getattr(oref, aname)
            if self.validate:
                try:
                    validate(msg, "%s-in.xsd" % (self.schemabase))
                except:
                    (trace, tval, trb) = exc_info()
                    syslog(LOG_ERR, "Validation In:%s:" % msg)
                    syslog(LOG_ERR, str(tval))
                    func = lambda x, y:Element("error", type='Validation', msg=str(tval))

            try:
                response = func(cmd, (peer, peerport))
            except:
                self.log_failure("command")
                response = Element("error", type='Semantic', msg='Unexpected Command Failure')
                syslog(LOG_ERR, "msg was: %s" % (msg))
        else:
            # Unhandled Command
            syslog(LOG_ERR, "unhandled cmd %s" % (cmd.tag))
            response = Element("error", type='Semantic', msg='%s is an unknown request' % cmd.tag)

        try:
            xmlresp = tostring(response)
        except:
            syslog(LOG_ERR, "Req was %s" % (msg))
            syslog(LOG_ERR, "Unexpected message send/serialize failure")
            self.ShutdownPeer(sock)
            return 1
                        
        # validate and send the response
        if self.validate:
            try:
                validate(xmlresp, "%s-out.xsd" % (self.schemabase))
            except ValidationException:
                (trace, val, trb) = exc_info()
                syslog(LOG_ERR, "Validation Out:%s:" % xmlresp)
                syslog(LOG_ERR, str(val))
                del trace, val, trb
                    
        try:
            self.comm.SendMessage(self.handles[sock], xmlresp)
        except (CommunicationError, ProtocolError, SocketError), msg:
            syslog(LOG_ERR, "%s from %s"%(msg, peer))
            self.ShutdownPeer(sock)
        except:
            syslog(LOG_ERR, "Unexpected message send/serialize failure")
            syslog(LOG_ERR, "Response.tagname=%s" % (response.tag))
            self.ShutdownPeer(sock)
        return 1


    def ParseConfig(self):
        '''Read config file'''
        self.cf = {}
        try:
            for data in open(self.cfname, 'r').readlines():
                (var, value) = data.split('=')
                self.cf[var] = value
        except IOError, err:
            syslog(LOG_ERR, "Unable to open conf file: ", err)

    def GetRefCount(self):
        '''Get reference counts for all objects'''
        data = {}
        # collect all classes
        for m in sys.modules.values():
            for sym in dir(m):
                o = getattr (m, sym)
                if type(o) is ClassType:
                    data[o] = sys.getrefcount (o)
        # sort by refcount
        pairs = [tuple(item) for item in data.items()]
        pairs.sort()
        pairs.reverse()
        return pairs
    
    def SaveState(self, depth = 0):
        '''Save important fields to a persistent store'''
        if depth > 3:
            raise StateSaveError
        state = tuple([getattr(self, field) for field in self.__statefields__])
        try:
            data = open(self.__statefile__, 'w')
            dump(state, data)
            data.close()
        except (IOError, OSError):
            raise StateSaveError
        except RuntimeError:
            self.SaveState(depth + 1)

    def LoadState(self):
        '''Load important fields from persistent store'''
        try:
            state = load(file(self.__statefile__))
        except:
            raise StateLoadError
        if len(state) != len(self.__statefields__):
            raise StateLoadError
        [setattr(self, self.__statefields__[idx], state[idx]) for idx in xrange(len(state))]

class EventReceiver(Server):
    '''The EventReceiver class is a server that automatically subscribes to event notifications.
    __subscriptions__ is a list of (c,m,d) tuples'''
    __implementation__ = "EventReceiver"
    __validate__ = 0
    __component__ = "er-%s-%s" % (getpid(), gethostname())
    __dispatch__ = {'events':'HandleEvent'}
    __subscriptions__ = []

    def HandleEvent(self, msg, (peer, port)):
        '''Null Event handler'''
        return Element("event-ok")

    def __postbind__(self):
        '''Handle subscription stuff'''
        msg = "<add-subscription><subscription component='%%s' msg='%%s' data='%%s' notify='%s'/></add-subscription>" % (self.__component__)
        try:
            handle = self.comm.ClientInit('event-manager')
            for sub in self.__subscriptions__:
                self.comm.SendMessage(handle, msg % sub)
                self.comm.RecvMessage(handle)
            self.comm.ClientClose(handle)
        except:
            self.log_failure("event subscription")
            raise SubscriptionException

    def __shutdown__(self):
        '''Shutdown subscriptions'''
        msg = "<del-subscription><subscription component='%%s' msg='%%s' data='%%s' notify='%s'/></del-subscription>" % (self.__component__)
        try:
            handle = self.comm.ClientInit('event-manager')
            for sub in self.__subscriptions__:
                self.comm.SendMessage(handle, msg % sub)
                self.comm.RecvMessage(handle)
            self.comm.ClientClose(handle)
        except:
            syslog(LOG_ERR, "Event Unsubscribe Failure.")

class Failure:
    '''The Failure class handles sparse reports of error conditions'''
    def __init__(self, name):
        self.name = name
        self.status = True

    def Fail(self):
        '''Set failure flag'''
        if self.status == True:
            syslog(LOG_ERR, "%s entered failure mode" % (self.name))
            self.status = False

    def Pass(self):
        '''Set pass flag'''
        if self.status == False:
            syslog(LOG_ERR, "%s cleared failure mode" % (self.name))
            self.status = True
